/* 
  Show the contents of any RIFF file.
  Orion Sky Lawlor, olawlor@acm.org, 2007/04/09 (Public Domain)
*/
#include <iostream>
#include <string>
#include <fstream> /* for ofstream */

typedef unsigned char io_byte;

class Lil32 { //Little-endian (Intel byte order) 32-bit integer
	io_byte d[4];
public:
	operator unsigned int () const { return d[0]|(d[1]<<8)|(d[2]<<16)|(d[3]<<24); }
};

/** Prints out contents of RIFF file */
class riff_printer {
public:
	std::istream &f;
	
	// Make a new class to deconstruct this file
	riff_printer(std::istream &f_) :f(f_),indent(0) {}
	// Print the contents of our RIFF file
	void print(void);
	
private:
	// Read the next 4 bytes as a RIFF tag
	std::string read_tag(void) {
		std::string tag(4,'?');
		f.read((char *)&tag[0],4);
		return tag;
	}
	// Read a length from the file
	int read_len(void) {
		Lil32 len;
		f.read((char *)&len,sizeof(len));
		return len;
	}
	// Print out the next 4 bytes as a RIFF tag
	std::string print_tag(const char *where) {
		std::string tag=read_tag();
		pl()<<"Tag "<<where<<" '"<<tag<<"'\n";
		return tag;
	}
	// Print out the next length
	int print_len(const char *where) {
		int len=read_len();
		pl()<<"Length "<<where<<" == "<<len<<" bytes\n";
		return len;
	}
	
	// Print out the next total_length bytes of the file as chunks.
	void print_chunks(int total_length);
	
	// Print out the next len bytes of the file as bytes.
	void print_data(int len);
	
	int indent;
	// Begin a new line for printing
	std::ostream &pl(void) {
		for (int i=0;i<indent;i++)
			std::cout<<"\t";
		return std::cout;
	}
};

// Print out the next bodyLength bytes of the file as bytes.
void riff_printer::print_data(int bodyLength) {
	indent++;
	
	// Figure out how much we should bother to print:
	const int maxLen=4*16;
	io_byte buf[maxLen];
	int len=bodyLength;
	if (len>maxLen) {len=maxLen;}
	f.read((char *)&buf,len);
	
	/* Print everything out in ascii (if possible) */
	for (int i=0;i<len;i++) {
		if (i%16==0) {if (i>0) printf("\n"); pl();}
		io_byte c=buf[i];
		if (c<32) 
			printf("-- ");
		else if (c>=127)
			printf("++ ");
		else 
			printf("%c  ",(char)c);
	}
	
	/* Print everything out in hex */
	for (int i=0;i<len;i++) {
		if (i%16==0) {printf("\n"); pl();}
		printf("%02X ",buf[i]);
	}
	printf("\n");
	
	/* If we skipped over some data, indicate this with ... */
	if (bodyLength>len) {
		int skipBytes=bodyLength-len;
		pl()<<"... skipping "<<skipBytes<<" bytes ...\n";
		f.seekg(skipBytes,std::ios_base::cur);
	}
	indent--;
}


// Print out the next total_length bytes of the file as chunks.
void riff_printer::print_chunks(int total_length)
{
	indent++; // Indent each chunk list from the previous one
	
	// Loop over chunks in file
	while (total_length>0) {
		// Print chunk header
		std::string tag=print_tag("of chunk");
		int len=print_len("of chunk");
		
		if (tag=="LIST") 
		{ // Chunk that contains other chunks
			print_tag("of sublist");
			print_chunks(len-4);
		} else { // Normal chunk--print data inside chunk
			print_data(len);
		}
		
		// Skip to the next chunk
		int chunk_length=8+len; /* tag, len, and data */
		total_length-=chunk_length;
	}
	
	indent--;
}

// Print out the whole file
void riff_printer::print(void)
{
	print_tag("at start of file");
	int len=print_len("of file");
	print_tag("for file type");
	print_chunks(len-4);
}

int main(int argc,char *argv[])
{
	const char *file="in.wav";
	if (argc>1) file=argv[1]; /* first argument is filename */
	std::ifstream f(file,std::ios_base::binary);
	if (!f) {std::cout<<"Sorry, couldn't open input file '"<<file<<"'\n"; exit(1);}
	riff_printer rp(f);
	rp.print();
	return 0;
}

